// SPDX-License-Identifier: GPL-2.0
/*
 * Non-cacheable heap device
 *
 * Copyright (c) 2022, ASML Netherlands B.V.
 *    Jeroen Vreeken <jeroen.vreeken@asml.com>
 */


#include <linux/module.h>
#include <linux/device.h>
#include <linux/cdev.h>
#include <linux/vmalloc.h>
#include <linux/fs.h>
#include <asm/set_memory.h>

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("Character device for mapping non cacheable memory to userspace");
MODULE_AUTHOR("Jeroen Vreeken <jeroen.vreeken@asml.com>");
MODULE_VERSION("20221014");


#ifdef pr_fmt
#undef pr_fmt
#endif
#define pr_fmt(fmt) KBUILD_MODNAME ": " fmt

static struct class *ncheap_class;
static dev_t ncheap_cdev_nr;
static struct device *ncheap_device;
static struct cdev *ncheap_cdev;

struct heap_entry {
	void *heap;
	unsigned long size;
	struct heap_entry *next;
};


static int ncheap_open(struct inode *inode, struct file *file)
{
	return nonseekable_open(inode, file);
}

static int ncheap_release(struct inode *inode, struct file *file)
{
	struct heap_entry *entry, *next;
	
	next = file->private_data;
	while (next) {
		unsigned long offset;

		entry = next;
		next = entry->next;

		for (offset = 0; offset < entry->size; offset += PAGE_SIZE) {
			set_memory_wb((unsigned long)entry->heap + offset, 1);
		}

		
		vfree(entry->heap);
		kfree(entry);
	}
	return 0;
}

static int ncheap_mmap(struct file *file, struct vm_area_struct *vma)
{
	int r = 0;
	unsigned long size = PAGE_ALIGN(vma->vm_end - vma->vm_start);
	void *new_heap;
	struct heap_entry *new_heap_entry;
	unsigned long offset = 0;

	pr_debug("mmap size: 0x%lx\n", size);
	new_heap = vmalloc_user(size);
	if (IS_ERR(new_heap)) {
		r = PTR_ERR(new_heap);
		pr_err("Failed to alloc heap: %d\n", r);
		goto err_heap;
	}	

	new_heap_entry = kzalloc(sizeof(struct heap_entry), GFP_KERNEL);
	if (IS_ERR(new_heap_entry)) {
		r = PTR_ERR(new_heap_entry);
		pr_err("Failed to alloc heap entry: %d\n", r);
		goto err_heap_entry;
	}
	
	new_heap_entry->heap = new_heap;
	new_heap_entry->size = size;
	new_heap_entry->next = file->private_data;
	file->private_data = new_heap_entry;

	for (offset = 0; offset < size; offset += PAGE_SIZE) {
		set_memory_uc((unsigned long)new_heap + offset, 1);
	}

	vma->vm_page_prot = pgprot_noncached(vma->vm_page_prot);
	r = remap_vmalloc_range(vma, new_heap, 0);
	if (r < 0) {
		pr_err("remap_vmalloc_range() failed: %d\n", r);
		goto err_remap_vmalloc_range;
	}

	return 0;

err_remap_vmalloc_range:
	kfree(new_heap_entry);
err_heap_entry:
	vfree(new_heap);
err_heap:
	return r;
}

static const struct file_operations ncheap_cdev_fops = {
	.owner   = THIS_MODULE,
	.open    = ncheap_open,
	.release = ncheap_release,
	.mmap    = ncheap_mmap,
};

static char *ncheap_devnode(struct device *dev, umode_t *mode)
{
	if (mode)
		*mode = 0666;;
	return NULL;
}

static int __init ncheap_init(void)
{
	int r = 0;
	
	pr_info("init\n");

	ncheap_class = class_create(THIS_MODULE, KBUILD_MODNAME);
	if (IS_ERR(ncheap_class)) {
		r = PTR_ERR(ncheap_class);
		pr_err("class_create() failed: %d\n", r);
		goto err_class_create;
	}
	ncheap_class->devnode = ncheap_devnode;
	
	
	r = alloc_chrdev_region(&ncheap_cdev_nr, 0, 1, KBUILD_MODNAME);
	if (r < 0) {
		pr_err("alloc_chrdev_region() failed: %d\n", r);
		goto err_alloc_chrdev_region;
	}

	ncheap_cdev = cdev_alloc();
	if (!ncheap_cdev) {
		r = PTR_ERR(ncheap_cdev);
		pr_err("cdev_alloc() failed: %d\n", r);
		goto err_cdev_alloc;
	}

	ncheap_cdev->ops = &ncheap_cdev_fops;
	ncheap_cdev->owner = THIS_MODULE;
	
	r = cdev_add(ncheap_cdev, ncheap_cdev_nr, 1);
	if (r < 0) {
		pr_err("cdev_add() failed: %d\n", r);
		goto err_cdev_add;
	}

	ncheap_device = device_create(ncheap_class, NULL, ncheap_cdev_nr, NULL, KBUILD_MODNAME);
	if (IS_ERR(ncheap_device)) {
		r = PTR_ERR(ncheap_device);
		pr_err("device_create() failed: %d\n", r);
		goto err_device_create;
	}

	return 0;

err_device_create:
err_cdev_add:
err_cdev_alloc:
	cdev_del(ncheap_cdev);
	unregister_chrdev_region(ncheap_cdev_nr, 1);	
err_alloc_chrdev_region:
	class_destroy(ncheap_class);
err_class_create:
	return r;
}


static void __exit ncheap_exit(void)
{
	device_destroy(ncheap_class, ncheap_cdev_nr);
	ncheap_device = NULL;

	cdev_del(ncheap_cdev);

	unregister_chrdev_region(ncheap_cdev_nr, 1);
	
	class_destroy(ncheap_class);

	pr_info("exit\n");
}

module_init(ncheap_init);
module_exit(ncheap_exit);
